export const INCIDENT_PACKET_PROFILE = "glyph.agent-incident-packet/0.1-draft" as const;
export const CONTROL_PLANE_PROFILE = "bulla.control-plane-alpha/0.1-experimental" as const;
export const MCP_SOURCE_PIN = "2026-07-28" as const;
export const MCP_SCHEMA_COMMIT =
  "271ecc9accafdd9b83a3c869fa67c22953b2af80" as const;
export const MCP_SCHEMA_SHA256 =
  "sha256:ef70b61f99b6d2e5e3b46863822eab08dff6a45bedc7a08914e0e5b133f40203" as const;
export const MCP_STANDARDS_GATE = "PASSED" as const;
export const PRODUCTION_STATUS =
  "BLOCKED_NO_REMOTE_MUTATION_AUTHORIZED" as const;
export const PRODUCTION_DEPLOYMENT_BLOCKED = "true" as const;
export const TRANSACTION_RECONCILER_BLOCKER =
  "BLOCKED_EXTERNAL_RECONCILER_NOT_IMPLEMENTED" as const;
export const KEY_ROTATION_RETIREMENT_BLOCKER =
  "BLOCKED_KEY_ROTATION_RETIREMENT_NOT_IMPLEMENTED" as const;
export const CAPABILITY_TOKEN_DERIVATION_VERSION =
  "bulla-control-plane-alpha-capability/hmac-sha256-v1" as const;
export const MCP_WIRE_PROTOCOL_REVISION = "2026-07-28" as const;
export const MCP_MEDIATED_VALUE = "alpha-mediated" as const;
